#!/usr/bin/python
""" NAME
        mem_multi.py

    SYNOPSIS
        python module

    DESCRIPTION
        Module to make memory mapped memory bunches or registers
        accesses on multiple PCIe40 board. 
        An enhancement of former mem.py.

    OPTIONS
        > ...

    EXAMPLE
        > ...

    AUTHOR
        PY Duval -- 2015

    lICENSE
    (c) Copyright CERN for the benefit of the LHCb Collaboration.
    Distributed under the terms of the GNU General Public Licence V3.
    ana.py script is free and open source software.

    CHANGELOG
        PY Duval 16/09/2016 adapt for multiboard usage
        PY Duval 07/11/2017 really made it multi boards        
"""
from ctypes import *
import sys
import getopt

# better to define LD_LIBRARY_PATH to point to the right libs directory
# ===================================================
# use a GLOBAL for library loading and init
# ===================================================
print "loading ", __name__
lli = CDLL("libpcie40.so")

# make provsion for 8 boards (like cubic crate usage)
LLI_MAX_BOARDS = 8
initDone = [False, False, False, False, False, False, False, False,
            False, False, False, False, False, False, False, False]
ecs = [-1, -1, -1, -1, -1, -1, -1, -1,
       -1, -1, -1, -1, -1, -1, -1, -1]

def init(board):
    global initDone
    global ecs
    ecs[board] = lli.ecs_open(board, 2)
    # print "open bar2 done on board {}...".format(board)
    initDone[board] = True


def read(board, add):
    """
    Reads at address add on board
    board : the board number
    add : address in memeory mapped area
    returns: a two values list
        status (-1 in case of error)
        read value

    NOTE: the mask (& 0xffffffff) is necessary to have the full
          unsigned value and not a negative one with high numbers
    """
    # print "mem read word at address ",hex(add & 0xffffffff)
    global initDone
    global ecs
    if initDone[board] is False:
        init(board)
    cval = c_int()
    status = lli.ecs_iordBar(board, 2, add, byref(cval))

    if ((cval.value & 0xffffffff) == 0xffffffff):
        print "ERROR: read 0xffffffff at "+hex(add & 0xffffffff)
        status = -1
    else:
        # print "OK passed read  at 0x%08x value 0x%08x" % \
        #       ((add & 0xffffffff),(cval.value & 0xffffffff))
        status = 0
    return (status, cval.value & 0xffffffff)


def write(board, add, val):
    """
    Write at address add
    board : the target board number
    add : address in memory mapped area
    val : value to write
    returns: a two values list
        status (-1 in case of error but may not be significant)
        read value
    NOTE: the mask (& 0xffffffff) is necessary to have th full \
          unsigned value and not a negative one with high numbers
    """
    # print "mem write ",hex(val & 0xffffffff),\
    #       " at address ",hex(add & 0xffffffff)
    global initDone
    global ecs
    if initDone[board] is False:
        init(board)
    cval = c_int(val)
    status = lli.ecs_iowrBar(board, 2, add, byref(cval))
    return (status, cval.value & 0xffffffff)


def readBloc(board, add, size, vals):
    """
    Reads a bloc of size words starting at address add
    board : the target board
    add : start address in memory mapped area
    size: number of words to read
    vals: a list of read data

    returns : status (0 success, -1 error)
    NOTE: the mask (& 0xffffffff) is necessary to have the full\
          unsigned value and not a negative one with high numbers
    """
    # print "mem read bloc of ",size," words at address ",hex(add & 0xffffffff)
    global initDone
    global ecs
    if initDone[board] is False:
        init(board)
    # read a bloc of contiguous words
    rvalArray = (c_int * size)()
    pt = pointer(rvalArray)

    status = lli.ecs_readBloc(board, 2, add, size, pt)
    # status = lli.ecs_readBlocPYD(board, 2, add, size, pt)
    for reg in pt.contents:
        # print "read bloc 0x%x" % (reg & 0xffffffff)
        vals.append(reg)
    return status


def writeBloc(board, add, size, vals):
    """
    Write a bloc of size words startting at address add
    board : the target board number
    add : address in memory mapped area
    size: the nb of words
    vals : list of values to write
    returns : status (O success, -1 error)

    NOTE: the mask (& 0xffffffff) is necessary to have the full\
          unsigned value and not a negative one with high numbers
    """
    # print "mem write bloc of ",size,\
    #       " words starting at",hex(add & 0xffffffff)
    global initDone
    global ecs

    if initDone[board] is False:
        init(board)
    rvalArray = (c_int * size)(*vals)
    pt = pointer(rvalArray)
#     pt = cast(rvalArray, POINTER(c_int * size))
    status = lli.ecs_writeBloc(board, 2, add, size, pt)
    # status = lli.ecs_writeBlocPYD(board, 2, add, size, pt)
    return(status)

def fillMemory(board, value):
      print('Fill Memory of board {} with {}'.format(board, value))
      startAddressMemory = 0x0
      currentAddressMemory = startAddressMemory
      endAddressMemory = 0xFFF
      avalonVal = value
      while currentAddressMemory < endAddressMemory:
         avalonAdd = currentAddressMemory
         write(board, avalonAdd, avalonVal)
         status,val = read(board, avalonAdd)
         if int(val) <> value:
            print('Error with fill {}'.format(value))
         currentAddressMemory = currentAddressMemory +4
      print "Done"

def main():
    # parse command line options
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h", ["help"])
    except getopt.error, msg:
        print msg
        print "for help use --help"
        sys.exit(2)
    for o, a in opts:
        if o in ("-h", "--help"):
            print "help"
            sys.exit(0)
    for arg in args:
        print arg  # process() is defined elsewhere

    # useless init()
    print read(0, 0x0)
    print write(0, 0x0, 0x33)
    print read(0, 0x0)

    print read(0, 0x4)
    print write(0, 0x4, 32)
    print read(0, 0x4)
    
    print write(0, 0x8, 0x88888888)

    data = [0x111, 0x222, 0x333, 0x444, 0x555]
    writeBloc(0, 0x0, len(data), data)
    got = []
    readBloc(0, 0x0, len(data), got)
    for val in got:
        print "0x%08x" % val
    
    fillMemory(0, 0xface)
    
if __name__ == "__main__":
    main()
